// Check an e-mail using S/MIME

#include "smime_check.h"

#include <bestiola/assert.h>
#include <bestiola/base64.h>
#include <exim4-signat/common.h>
#include <nspr/nspr.h>
#include <nss/cert.h>
#include <nss/cms.h>
#include <nss/nss.h>
#include <nss/secerr.h>

static int skip_header(FILE *stream);
static bool update_context(FILE *stream, const char *mm_boundary,
	NSSCMSDigestContext *cmsdigcx);
static bool update_decoder(FILE *stream, NSSCMSDecoderContext *p7dcx);

static bool check_main_value(FILE *stream, const char *value) {
	int c = getc(stream);
	while (c == '\t' || c == ' ')
		c = getc(stream);
	while (c != EOF) {
		if (*value) {
			char lower_c = to_c_lower(c);
			if (lower_c != *value++)
				return false;
		}
		else {
			switch (c) {
			case '\n':
				if (ungetc(c, stream) == EOF)
					return false;
			case '\t':
			case '\r':
			case ' ':
			case ';':
				return true;
			}
			return false;
		}
		c = getc(stream);
	}
	return false;
}

static bool decode_signature(FILE *stream, const char *sender,
	SECOidTag digestalgtag, SECItem *digest) {
	NSSCMSDecoderContext *p7dcx = NSS_CMSDecoder_Start(NULL, NULL, NULL,
		NULL, NULL, NULL, NULL);
	if (!p7dcx)
		return true;
	if (update_decoder(stream, p7dcx)) {
		NSS_CMSDecoder_Cancel(p7dcx);
		return true;
	}
	NSSCMSMessage *cmsg = NSS_CMSDecoder_Finish(p7dcx);
	if (!cmsg)
		return true;

	NSSCMSContentInfo *cinfo;
	int count = NSS_CMSMessage_ContentLevelCount(cmsg);
	int n;
	for (n = 0; n < count; n++) {
		cinfo = NSS_CMSMessage_ContentLevel(cmsg, n);
		if (NSS_CMSContentInfo_GetContentTypeTag(cinfo)
			== SEC_OID_PKCS7_SIGNED_DATA)
			break;
	}
	if (!(n < count)) {
		NSS_CMSMessage_Destroy(cmsg);
		return true;
	}

	NSSCMSSignedData *sigd = NSS_CMSContentInfo_GetContent(cinfo);
	if (!sigd || NSS_CMSSignedData_HasDigests(sigd)) {
		NSS_CMSMessage_Destroy(cmsg);
		return true;
	}

	CERTCertDBHandle *certdb = CERT_GetDefaultCertDB();
	if (NSS_CMSSignedData_SetDigestValue(sigd, digestalgtag, digest)
		!= SECSuccess
		|| NSS_CMSSignedData_ImportCerts(sigd, certdb,
		certUsageEmailSigner, PR_FALSE) != SECSuccess) {
		NSS_CMSMessage_Destroy(cmsg);
		return true;
	}

	if (NSS_CMSSignedData_SignerInfoCount(sigd) < 1) {
		NSS_CMSMessage_Destroy(cmsg);
		return true;
	}

	NSSCMSSignerInfo *sinfo = NSS_CMSSignedData_GetSignerInfo(sigd, 0);
	CERTCertificate *cert = NSS_CMSSignerInfo_GetSigningCertificate(sinfo,
		NULL);
	if (!cert) {
		NSS_CMSMessage_Destroy(cmsg);
		return true;
	}
	char *emailAddr = cert->emailAddr;
	if (!emailAddr || strcmp(sender, emailAddr)) {
		NSS_CMSMessage_Destroy(cmsg);
		return true;
	}

	if (NSS_CMSSignedData_VerifySignerInfo(sigd, 0, certdb,
		certUsageEmailSigner) != SECSuccess) {
		NSS_CMSMessage_Destroy(cmsg);
		return true;
	}

	NSS_CMSMessage_Destroy(cmsg);
	return false;
}

static bool digest_text(FILE *stream, const char *mm_boundary, SECOidTag tag,
	PLArenaPool *poolp, SECItem *digest) {
	PLArenaPool *arena = PORT_NewArena(1024);
	if (!arena)
		return true;
	SECAlgorithmID digestalg;
	if (SECOID_SetAlgorithmID(arena, &digestalg, tag, NULL)
		!= SECSuccess) {
		PORT_FreeArena(arena, PR_FALSE);
		return true;
	}
	NSSCMSDigestContext *cmsdigcx
		= NSS_CMSDigestContext_StartSingle(&digestalg);
	if (!cmsdigcx) {
		PORT_FreeArena(arena, PR_FALSE);
		return true;
	}

	if (update_context(stream, mm_boundary, cmsdigcx)) {
		NSS_CMSDigestContext_Cancel(cmsdigcx);
		PORT_FreeArena(arena, PR_FALSE);
		return true;
	}

	bool error = NSS_CMSDigestContext_FinishSingle(cmsdigcx, poolp, digest)
		!= SECSuccess;
	PORT_FreeArena(arena, PR_FALSE);
	return error;
}

static SECOidTag get_micalg_tag(const char *micalg) {
	static struct {
		const char *micalg;
		SECOidTag tag;
	} table[] = {
		{ "sha-256", SEC_OID_SHA256 },
		{ "sha1", SEC_OID_SHA1 }
	};

	unsigned i;
	for (i = 0; i < sizeof table / sizeof table[0]; i++) {
		if (!strcmp(micalg, table[i].micalg))
			return table[i].tag;
	}
	return SEC_OID_UNKNOWN;
}

static bool scan_headers(FILE *stream) {
	bool good_encoding = false;
	bool good_type = false;
	int c = getc(stream);
	for (;;) {
		switch (c) {
		case '\r':
			c = getc(stream);
			if (c != '\n')
				return true;
		case '\n':
			return !(good_encoding && good_type);
		case EOF:
			return true;
		}
		int i = 0;
		bool maybe_encoding = true;
		bool maybe_type = true;
		for (;;) {
			char lower_c = to_c_lower(c);
			if (maybe_encoding) {
				static const char ENCODING[]
					= "content-transfer-encoding";
				if (ENCODING[i]) {
					if (lower_c != ENCODING[i])
						maybe_encoding = false;
				}
				else if (c == ':') {
					if (good_encoding
						|| !check_main_value(stream,
						"base64"))
						return true;
					good_encoding = true;
					break;
				}
				else
					maybe_encoding = false;
			}
			if (maybe_type) {
				static const char TYPE[] = "content-type";
				if (TYPE[i]) {
					if (lower_c != TYPE[i])
						maybe_type = false;
				}
				else if (c == ':') {
					if (good_type
						|| !check_main_value(stream,
						"application/pkcs7-signature"))
						return true;
					good_type = true;
					break;
				}
				else
					maybe_type = false;
			}
			if (!(maybe_encoding || maybe_type))
				break;
			i++;
			c = getc(stream);
			if (c == EOF)
				return true;
		}
		c = skip_header(stream);
	}
	return true;
}

static int skip_header(FILE *stream) {
	int c = getc(stream);
	for (;;) {
		switch (c) {
		case '\n':
			c = getc(stream);
			if (c == '\t' || c == ' ')
				break;
		case EOF:
			return c;
		}
		c = getc(stream);
	}
}

bool smime_check(const char *sender, FILE *stream, const char *micalg,
	const char *mm_boundary, const char *configdir) {
	SECOidTag micalg_tag = get_micalg_tag(micalg);
	if (micalg_tag == SEC_OID_UNKNOWN)
		return true;

	PR_Init(0, 0, 0);
	if (NSS_Init(configdir) != SECSuccess) {
		PR_Cleanup();
		return true;
	}
	PLArenaPool *poolp = PORT_NewArena(1024);
	if (!poolp) {
		NSS_Shutdown();
		PR_Cleanup();
		return true;
	}

	SECItem digest;
	bool error = digest_text(stream, mm_boundary, micalg_tag, poolp,
		&digest) || scan_headers(stream)
		|| decode_signature(stream, sender, micalg_tag, &digest);

	PORT_FreeArena(poolp, PR_FALSE);
	if (NSS_Shutdown() != SECSuccess)
		error = true;
	if (PR_Cleanup() != PR_SUCCESS)
		error = true;
	return error;
}

static bool update_context(FILE *stream, const char *mm_boundary,
	NSSCMSDigestContext *cmsdigcx) {
	if (BUFSIZ < 80)
		bad_assertion();
	unsigned char *data = malloc(BUFSIZ);
	if (!data)
		return true;
	int len = 0;

	int state = 0;
	int i = 0;
	int c = getc(stream);
	while (c != EOF) {
		switch (state) {
		case 0:
			switch (c) {
			case '\n':
				state = 2;
				break;
			case '\r':
				state++;
				break;
			default:
				if (len + 1 > BUFSIZ) {
					NSS_CMSDigestContext_Update(cmsdigcx,
						data, len);
					len = 0;
				}
				data[len++] = c;
			}
			break;
		case 1:
			if (c == '\n')
				state++;
			else {
				state = 0;
				if (len + 2 > BUFSIZ) {
					NSS_CMSDigestContext_Update(cmsdigcx,
						data, len);
					len = 0;
				}
				data[len++] = '\r';
				data[len++] = '\n';
				continue;
			}
			break;
		case 2:
			if (c == mm_boundary[i]) {
				if (!mm_boundary[++i])
					state++;
			}
			else {
				state = 0;
				if (len + 2 + i > BUFSIZ) {
					NSS_CMSDigestContext_Update(cmsdigcx,
						data, len);
					len = 0;
				}
				data[len++] = '\r';
				data[len++] = '\n';
				if (i) {
					memcpy(data + len, mm_boundary, i);
					len += i;
					i = 0;
				}
				continue;
			}
			break;
		case 3:
			if (c == '\r') {
				state++;
				break;
			}
		case 4:
		{
			bool error = c != '\n';
			if (!error)
				NSS_CMSDigestContext_Update(cmsdigcx, data,
					len);
			free(data);
			return error;
		}
		}
		c = getc(stream);
	}
	free(data);
	return true;
}

static bool update_decoder(FILE *stream, NSSCMSDecoderContext *p7dcx) {
	char *buf = malloc(BUFSIZ);
	if (!buf)
		return true;
	unsigned long len = 0;

	bool last = false;
	int i;
	while (!last) {
		char b[4];
		i = 0;
		for (;;) {
			int c = getc(stream);
			if (c == EOF) {
				free(buf);
				return true;
			}
			if (is_base64(c)) {
				b[i] = c;
				if (i++ == 3)
					break;
			}
			else if (c == '-' || c == '=') {
				if (i == 1) {
					free(buf);
					return true;
				}
				last = true;
				int j;
				for (j = i; j < 4; j++)
					b[j] = 'A';
				break;
			}
		}

		if (len + 3 > BUFSIZ) {
			if (NSS_CMSDecoder_Update(p7dcx, buf, len)
				!= SECSuccess) {
				free(buf);
				return true;
			}
			len = 0;
		}

		unsigned char v[4];
		int j;
		for (j = 0; j < 4; j++)
			v[j] = base64_value(b[j]);
		buf[len++] = v[0] << 2 | v[1] >> 4;
		buf[len++] = v[1] << 4 | v[2] >> 2;
		buf[len++] = v[2] << 6 | v[3];
	}

	switch (i) {
	case 0:
		break;
	case 2:
		len -= 2;
		break;
	case 3:
		len--;
	}
	if (NSS_CMSDecoder_Update(p7dcx, buf, len) != SECSuccess) {
		free(buf);
		return true;
	}
	free(buf);
	return false;
}
